import yaml from 'js-yaml';

export const COLOR_BY_TYPES = ['instance', 'activity'];

/* eslint-disable */
export const ACTIVITY_COLORS = {
  'Casing_Facility': '#50EBEC',
  'Carry_on_Back': 'red',
  'Enter_Facility': 'orange',
  'Exit_Facility': 'yellow',
  'Hand_Interaction': 'green',
  'Joining_Queue': '#E38AAE',
  'Object_Transfer': 'lime',
  'Open_Facility_Door': 'cyan',
  'People_Talking': 'cyan',
  'Pick_Up_Object': 'lime',
  'Purchasing': 'yellow',
  'Read_Document': 'orange',
  'Set_Down_Object': 'lightblue',
  'Sit_Down': 'magenta',
  'Stand_Up': 'magenta',
  'Text_On_Phone': '#F75D59',
  'Unload_Vehicle': 'magenta',
  'Unknown': 'lime',
  'Vehicle_Moving': 'yellow',
  'default': 'yellow',
  'input': 'red',
  'output': 'white',
}; /* eslint-enable */

export const GEOM_ACTOR_KEY = 'id1';
export const TIMERANGE_KEY = 'tsr0';

/**
 * @param {File} inputFile
 * @returns {Promise} stringPromise
 */
const readUploadedFileAsText = (inputFile, fallback) => {
  const temporaryFileReader = new FileReader();
  return new Promise((resolve, reject) => {
    temporaryFileReader.onerror = (err) => {
      console.error(err);
      temporaryFileReader.abort();
      reject(new DOMException('Problem parsing input file.'));
    };
    temporaryFileReader.onload = () => {
      resolve(temporaryFileReader.result);
    };
    try {
      temporaryFileReader.readAsText(inputFile);
    } catch(err) {
      if (fallback) {
        resolve(fallback);
      } else {
        throw err;
      }
    }
  });
};

/**
 * KPF Class parses annotations defined in KPF yaml format.
 */
export default class KPF {

  constructor(activities = [], geom = [], types = [], meta = {}) {
    this.activities = activities;
    this.geom = geom;
    this.types = types;
    this.tracksByActor = {};
    this.meta = meta;
    this.postProcess();
  }

  /**
   * postProcess:
   * - convert g0 from string to array, call it bbox
   * - make sure ts0 is present on detections
   * - remove meta
   * - add additional properties:
   *   + activity.actor[i].detections (time-ordered)
   *   + activity.actor[i].type
   *   + activity.name
   */
  postProcess() {
    this.geom = KPF.filterEmpty(this.geom, [ GEOM_ACTOR_KEY, 'g0', 'ts0' ]);
    this.activities = KPF.filterEmpty(this.activities, [ 'act2' ]);
    this.types = KPF.filterEmpty(this.types, [GEOM_ACTOR_KEY]);
    const newTypes = {};
    this.types.forEach(t => {
      newTypes[t[GEOM_ACTOR_KEY]] = {
        id: t.id,
        type: Object.keys(t.cset3)[0] || "Unknown",
      };
    });
    this.types = newTypes;
    this.geom = this.geom.map(g => ({
      ...g,
      bbox: g.bbox || g.g0.split(' ').map(f => parseFloat(f)),
      ts0: g.ts0 || g.frame,
      [GEOM_ACTOR_KEY]: g[GEOM_ACTOR_KEY] || 0,
    }));
    this.geom.forEach((g) => {
      const key = g[GEOM_ACTOR_KEY];
      if (key in this.tracksByActor) {
        this.tracksByActor[key].push(g);
      } else {
        this.tracksByActor[key] = [g];
      }
    });
    this.activities = this.activities.map(a => ({
      ...a,
      actors: a.actors.map((actor) => ({
        ...actor,
        timespan: actor.timespan[0][TIMERANGE_KEY],
        detections: KPF.filterRange(
            this.tracksByActor[actor[GEOM_ACTOR_KEY]],
            'ts0',
            actor.timespan[0][TIMERANGE_KEY][0],
            actor.timespan[0][TIMERANGE_KEY][1],
        ),
        type: this.types[actor[GEOM_ACTOR_KEY]]
          ? this.types[actor[GEOM_ACTOR_KEY]].type
          : 'Unknown',
      })),
      name: a.act2 ? Object.keys(a.act2)[0] : a.src, // TODO: this is terrible.
    }));
  }

  /**
   * return a list of the tracks
   * @returns {Array.<track>}
   */
  get tracks() {
    return [].concat(...this.activities.map(activity =>
      activity.actors.map(actor => ({
        activity,
        actor,
        timespan: actor.timespan,
        detections: actor.detections,
        actorColor: actor.randomColor,
      })),
    )).sort((a, b) => {
      return a.detections[0].ts1 - b.detections[0].ts1;
    });
  }

  /**
   * Return elements of list that have validKey defined
   * @param {Array} list
   * @param {String} validKey
   */
  static filterEmpty(list, validKeyList) {
    return list.filter(l => validKeyList.some(e => e in l));
  }

  /**
   * Return elements where number key in range value
   */
  static filterRange(list, key, lower, upper) {
    return list.filter(item => item[key] <= upper && item[key] >= lower);
  }

  /**
   * Create KPF object from YAML text strings.
   * @param {String} activityText
   * @param {String} geometryText
   */
  static fromText(activityText, geometryText, typeText) {
    const activityJson = yaml.safeLoad(activityText).map(a => ({ ...a.act }));
    const geometryJson = yaml.safeLoad(geometryText).map(g => ({ ...g.geom }));
    const typeJson = yaml.safeLoad(typeText).map(t => ({ ...t.types }));
    return new KPF(activityJson, geometryJson, typeJson);
  }

  /**
   * Create KPF from js file handles.
   * @param {File} activityFile
   * @param {File} geometryFile
   */
  static async fromFiles(activityFile, geometryFile, typeFile) {
    const promiseList = [
      readUploadedFileAsText(activityFile),
      readUploadedFileAsText(geometryFile),
      readUploadedFileAsText(typeFile, '- {}'),
    ];
    const [activityText, geometryText, typeText] = await Promise.all(promiseList);
    return KPF.fromText(activityText, geometryText, typeText);
  }
}
