/*

TDM is a format for visualizing annotations on video.
It is a lossy format, and should not be used to store or
save information about an annotation. TDM is designed to
be simple enough that any annotation generation method
should produce sufficient data to transcribe to TDM.

TDM stands for Tracks, Detections, and Metadata.

Some metadata properties are considered first-class, such
as "color" and "label".  They can directly affect how the track
is rendered.  Other metadata properties can be tertiary, and
exist for the purpose of easily finding and manipulating other
properties.

For example, "actor_id" has no meaning in TDM, but a client
application may want to set "state" = "hidden" for all "actor_id"
matching a certain value.

*/

export const SORT_RESOLUTION = 50;
export const CONTRAST_COLORS = [
  "#FF00FF", // "fuchsia",
  "#FFFF00", // "yellow",
  "#FF0000", // "red",
  "#00FFFF", // "aqua",
  "#00FF00", // "lime",
  "#6A5ACD", // "slateblue",
  "#FF69B4", // "hotpink",
];
export const SEQUENCE_COLORS = [
  '#543005',
  '#8c510a',
  '#bf812d',
  '#dfc27d',
  '#f6e8c3',
  '#c7eae5',
  '#80cdc1',
  '#35978f',
  '#01665e',
  '#003c30',
];
export const DISABLED_COLOR = '#a0a0a0';

/**
 * @typedef {string} TrackState
 */

/**
 * Enum for track state
 * @readonly
 * @enum {TrackState}
 */
export const STATES = {
  ACTIVE: "active",
  DISABLED: "disabled",
  HIDDEN: "hidden",
};

/**
 * @typedef {number} ShapeType
 */

/**
 * Enum for shape types
 * @readonly
 * @enum {ShapeType}
 */
export const SHAPES = {
  POINT: 1,
  LINE: 2,
};

/**
 * Meta Schema
 * @typedef {Obect} Metadata
 * ... Can have any properties.
 */

/**
 * Shape Schema
 * @typedef {Object} Shape
 * @property {any} data
 * @property {ShapeType} type
 */

/**
 * Detection Schema
 * @typedef {Object} Detection
 * @property {Number} frame
 * @property {Array<Number>} box [x_left, y_top, x_right, y_bottom]
 * @property {Track} track
 * @property {Metadata} meta
 */

/**
 * Track Schema
 * @typedef {Object} Track
 * @property {Metadata} meta
 * @property {Array<Detection>} detections
 * @property {Number} begin
 * @property {Number} end
 * @property {Object.<number, Array<Detection>>} sorted
 * @property {Boolean} interpolated whether the detections need interpolation
 * @property {String} color html color string
 * @property {Array<Shapes>} shapes
 * @property {TrackState} state
 */

/**
 * TDM Schema
 * @typedef {Object} TDM
 * @property {Object} proxy for accessing tracks by proxy.tracks
 * @property {Number} refresh for recording when changes happen to the track object.
 */

/**
 * Create a TDM object where tracks are hidden behind a proxy.
 * @param {Array<Track>} tracks
 * @param {Number} refresh
 * @returns {TDM}
 */
function makeTdm(tracks, refresh) {
  const handler = {
    get: (obj, prop) => {
      if (prop === 'tracks') {
        return tracks;
      }
    },
  };
  const p = new Proxy({}, handler);
  return {
    refresh,
    proxy: p,
  };
}

/**
 * Sort detections for a given track into buckets based on resolution size.
 * @param {Array<Track>} tracks
 * @param {Number} resolution in frames
 * @returns {Array<Track>}
 */
function bucketSort(tracks, resolution = SORT_RESOLUTION) {
  tracks.forEach(track => {
    const bucket = {};

    track.detections.forEach(d => {
      const key = Math.floor(d.frame / resolution) * resolution;
      if (key in bucket) {
        bucket[key].push(d);
      } else {
        bucket[key] = [d];
      }
    });

    for (let k in bucket) {
      bucket[k] = bucket[k].sort((a, b) => a.frame - b.frame);
    }

    track.sorted = bucket;
  });
  return tracks;
}

/**
 * Returns uninterpolated active frames.
 * @param {Array<Track>} tracks
 * @param {Number} frametime
 * @returns {Array<Detection>}
 */
function getActiveContinuousFrames(tracks, frametime) {
  const continuousFrames = [];
  const f = Math.floor(frametime);
  const key = Math.floor(f / SORT_RESOLUTION) * SORT_RESOLUTION;
  tracks.forEach((track) => {
    if (!track.interpolated && key in track.sorted) {
      const detections = track.sorted[key];
      const firstDetectionTime = detections[0].frame;
      const detectionIndex = f - firstDetectionTime;
      if (detectionIndex >= 0 && detections.length > detectionIndex) {
        continuousFrames.push(detections[detectionIndex]);
      }
    }
  });
  return continuousFrames;
}

/**
 * Returns interpolated frames from active keyframe interval.
 * @param {Array<Track>} tracks
 * @param {Number} frametime
 * @returns {Array<Detection>}
 */
function getActiveKeyFrames(tracks, frametime) {
  const frames = tracks
    .filter(trackinfo => trackinfo.interpolated)
    .map((trackinfo) => {
      const track = trackinfo.detections;
      for (let i = 0; i < track.length - 1; i += 1) {
        const d0 = track[i];
        const d1 = track[i + 1];
        if (frametime >= d0.frame && frametime <= d1.frame) {
          return interpolateFrames(frametime, d0, d1);
        }
      }
      return null;
    })
    .filter(track => track !== null);
  return frames;
}

/**
 * Returns the linear interpolation between d0 and d1 at currentFrame.
 * @param {Number} currentFrame
 * @param {Detection} d0
 * @param {Detection} d1
 * @returns {Detection}
 */
function interpolateFrames(currentFrame, d0, d1) {
  const len = d1.frame - d0.frame;
  // a + b = 1; interpolate from a to b
  const b = Math.abs((currentFrame - d0.frame) / len);
  const a = 1 - b;
  const box = d0.box.map((_, i) => ((d0.box[i] * a) + (d1.box[i] * b)));
  const frame = Math.round((d0.frame * a) + (d1.frame * b));
  return { ...d0, frame, box };
}

/**
 * @param {Array<Number>} box
 * @param {Number} scale (0, 1]
 * @returns {Array<Number>}
 */
function scaleBox(box, scale) {
  return box.map(n => (n * scale));
}

/**
 * Group tracks by meta key
 * @param {TDM} tdm
 * @param {String} key
 * @returns {Object<String, String>} value to color map
 */
function colorByMeta(tdm, key) {
  const assignedColors = {};
  let counter = 0;
  tdm.proxy.tracks.forEach(t => {
    const val = t.meta[key];
    if (val in assignedColors) {
      t.color = assignedColors[val];
    } else {
      const color = CONTRAST_COLORS[counter % CONTRAST_COLORS.length];
      t.color = color
      assignedColors[val] = color;
      counter += 1;
    }
  });
  tdm.refresh += 1;
  return assignedColors;
}

/**
 * Color tracks based on values in an array that interpolate from a to b.
 * @param {TDM>} tdm
 * @param {String} key key to a meta-property of track with value Z ∈ [0, 100]
 * @returns {Object<String, String>} value to color map.
 */
function colorByBucketSequence(tdm, key) {
  const assignedColors = {};
  tdm.proxy.tracks.forEach(t => {
    const value = t.meta[key] || -1;
    const i = Math.floor(value / SEQUENCE_COLORS.length);
    const color = i >= 0 ? SEQUENCE_COLORS[i] : DISABLED_COLOR;
    assignedColors[value] =  color;
    t.color = color;
  });
  tdm.refresh += 1;
  return assignedColors;
}

/**
 * Enable or disable tracks by meta key
 * @param {TDM} tdm
 * @param {String} key
 * @param {Any} value
 * @param {Boolean} newState state to set iff key = value
 * @param {Boolean} elseState state to set iff key != value
 * @param {Boolean} xor whether or not to set elseState if else condition
 * @returns {Object<String, Boolean>} value to enabled map
 */
function toggleByMeta(tdm, key, value,
  newState = STATES.DISABLED,
  elseState = STATES.ACTIVE,
  xor = false) {

  const enabledMap = {};
  tdm.proxy.tracks.forEach(t => {
    if (String(t.meta[key]) === value) {
      t.state = newState;
      enabledMap[t.meta[key]] = newState;
    } else if (xor) {
      t.state = elseState;
      enabledMap[t.meta[key]] = elseState;
    } else {
      enabledMap[t.meta[key]] = t.state;
    }
  });
  tdm.refresh += 1;
  return enabledMap;
}

/**
 * Reset all tracks to enabled.
 * @param {TDM} tdm
 * @param {String} key
 * @returns {Object<String, Boolean>} value to enabled map
 */
function resetState(tdm, key, newstate = STATES.ACTIVE) {
  const enabledMap = {}
  tdm.proxy.tracks.forEach(t => {
    t.state = newstate;
    enabledMap[t.meta[key]] = newstate;
  })
  tdm.refresh += 1;
  return enabledMap;
}

/**
 * Filter some item with meta-properties by key/val.
 * @param {Array<Track>|Array<Detection>} items
 * @param {String} key
 * @param {Array} values
 * @returns {Array<Track>|Array<Detection>}
 */
function filterByMeta(items, key, values) {
  return items.filter(v => values.indexOf(v.meta[key]) >= 0);
}

/**
 * Given a rectangle and two unit vectors,
 * get the point in the rectangle at (x, y)
 * @param {Array<Number>} box
 */
function centroid(box, x = 0.5, y = 0.5) {
  return [
    Math.round((box[2] * x) + (box[0] * x)),
    Math.round((box[3] * y) + (box[1] * y)),
  ];
}

/**
 * Generate a list of centroids for a given track.
 * @param {Array<Detection>} detections
 * @returns {Array<Shape>}
 */
function generateCentroids(detections) {
  return detections.map(d => ({
    type: SHAPES.POINT,
    data: {
      point: centroid(d.box),
      radius: 4, // unitless scalar
    },
  }));
}

/**
 * Generate a list of lines between all centroids for a track.
 * @param {Array<Detection>} detections
 * @returns {Array<Shape>}
 */
function generateDetectionLines(detections, x = 0.5, y = 0.5) {
  return detections.slice(0, detections.length - 1).map((_, i) => ({
    type: SHAPES.LINE,
    data: {
      width: 2,
      a: centroid(detections[i].box, x, y),
      b: centroid(detections[i+1].box, x, y),
    },
  }));
}

export {
  bucketSort,
  centroid,
  colorByBucketSequence,
  colorByMeta,
  filterByMeta,
  generateCentroids,
  generateDetectionLines,
  getActiveContinuousFrames,
  getActiveKeyFrames,
  interpolateFrames,
  makeTdm,
  resetState,
  scaleBox,
  toggleByMeta
};
