import KPF, { GEOM_ACTOR_KEY } from './kpf';
import { STATES, bucketSort } from './tdm';

function boxFromDetection(d) {
  return [d.left, d.top, d.right, d.bottom];
}

/**
 * Turn refiner endpint data into TDM tracks.
 * @param {Object} data raw data from refiner endpoint
 * @param {Object} staticMeta static metadata to assign to all tracks.
 * @returns {Array<Track>}
 */
function fromRefiner(data, staticMeta = {}) {
  return bucketSort(data.tracks.map((t, i) => {
    const meta = Object.assign({
      name: data.type,
      type: t.type,
      track_index: t.track_index,
      id: t.id,
      actor_id: i + 1,
      activity_id: 1,
    }, staticMeta);
    const track = {
      meta,
      interpolated: true,
      color: 'yellow',
      state: STATES.ACTIVE,
      begin: data.activity.start_frame,
      end: data.activity.end_frame,
    };
    const detections = t.detections.map(d => ({
      frame: d.frame,
      meta: {
        src: d.source,
        occlusion: d.occlusion,
      },
      box: boxFromDetection(d),
      track,
    })).sort((a, b) => a.frame - b.frame);
    track.detections = detections;
    return track;
  }));
}

/**
 * Turn KPF into TDM tracks.
 * @param {KPF} kpf object
 * @param {Object} staticMeta static metadata to assign to all tracks.
 * @returns {Array<Track>}
 */
function fromKPF(kpf, staticMeta = {}) {
  return bucketSort(kpf.tracks.map(t => {
    const meta = Object.assign({
      ...kpf.meta,
      name: t.activity.name,
      activity_id: t.activity.id2,
      type: t.actor.type,
      id: t.activity.id,
      actor_id: t.actor[GEOM_ACTOR_KEY],
      src_status: t.activity.src_status || 'unset',
    }, staticMeta);
    const track = {
      meta,
      interpolated: t.detections.length < (t.timespan[1] - t.timespan[0]),
      color: 'green',
      state: STATES.ACTIVE,
      begin: t.timespan[0],
      end: t.timespan[1],
    };
    const detections = t.detections.map(d => ({
      frame: d.frame || d.ts0,
      meta: {
        src: d.src,
        occlusion: d.occlusion,
      },
      box: d.bbox,
      track,
    })).sort((a, b) => a.frame - b.frame);
    track.detections = detections;
    return track;
  }));
}


export {
  fromKPF,
  fromRefiner,
};
